#Generic communication functions

import os
import os.path
import types
import time
import signal
import sys
import fcntl
import struct
import array
import subprocess

from socket import socket, AF_INET, SOCK_DGRAM

import gwutil
from serverdetails import *
from variables import *
import gwschedule
from gwlogging import logger, gw_stdout, gw_stderr


def chooseServer(servers, gprs=False):
    """Cycles through the dictionary of servers and returns the name
    of first one which can be sucessfully pinged otherwise returns none"""
    logger.debug("Chooose server")
    logger.info("GPRS connection %s" % gprs)
    for name, server in servers.items():
        #check if server is only accesible via any connection, if so try ping
        # or if only available via gprs check using gprs connection
        if (not server.gprs_only
                or (gprs and server.gprs_only)):
            cmd = "ping -c 3 %s" % server.address
            logger.debug(cmd)
            p = subprocess.Popen(args=cmd,
                                    shell=True,
                                    stdout=gw_stdout,
                                    stderr=gw_stderr)
            rc = p.wait()
            if rc == 0:
                logger.debug("Server chosen: %s" % name)
                return name
    return None


def chooseWebServer(sshServer=None):
    """Checks to see if the chosen ssh server is also a valid web server,
    if it is then use that, if not find a valid server"""
    if sshServer is not None and sshServer in webServers.keys():
        logger.debug("Using same server as SSH for web")
        return sshServer
    return chooseServer(webServers)


def chooseSshServer(webServer=None, gprs=False):
    """Checks to see if the chosen web server is also a valid ssh server,
    if it is then use that, if not find a valid server"""
    if (webServer is not None
            and webServer in sshServers.keys()
            and (gprs or sshServers[webServer].gprs_only == False)):
        logger.debug("using web server for ssh")
        return webServer
    return chooseServer(sshServers, gprs)


def getSpecial(server_name, mode):
    """Gets the file from the specified server"""
    remote_special = "/%s_special" % (mode)
    logger.debug("Remote file = %s" % remote_special)
    server = webServers[server_name]
    url = server.address + server.webPath + remote_special
    logger.info("Fetching special file from %s" % url)
    wget = subprocess.Popen(args="/usr/bin/wget -q  -O %s %s" % (SPECIAL_FILE, url),
                    shell=True,
                    stdout=gw_stdout,
                    stderr=gw_stderr)
    wget.wait()


def sendPowerState(server_name, mode, state):
    server = webServers[server_name]
    # Send our power state home
    logger.info("Sending power state home")
    wget = subprocess.Popen(
                args="/usr/bin/wget -q -O /dev/null -q %s%s%s?id=%s\&state=%i" % (
                server.address, server.webPath, POWER_UPLOAD, mode, state),
                shell=True,
                stdout=gw_stdout,
                stderr=gw_stderr)
    wget.wait()


def fetchOverridePowerState(server_name, mode):
    server = webServers[server_name]
    # Get override power state
    logger.info("Fetching overide power state")
    wget = subprocess.Popen(args="/usr/bin/wget -O %s -q %s%s%s?id=%s" % (
                            OVERIDE_FILE, server.address, server.webPath,
                            POWER_DOWNLOAD, mode),
                shell=True,
                stdout=gw_stdout,
                stderr=gw_stderr)
    wget.wait()


def uploadFiles(server_name, force_all=True, connection=None):
    """ Upload the files that are queued to the server specified"""
    server = sshServers[server_name]
    # Copy  the tars across:
    logger.info("Sending queued files to %s:%s" % (server.address,
        server.uploadPath))
    for fname in os.listdir(BASE_DIR + QUEUED_DIR):
        size = gwutil.file_size(BASE_DIR + QUEUED_DIR + "/" + fname)
        if (force_all) or (connection == "WIFI") or (size < SMALL_FILE_SIZE):
            # Copy the data:
            cmd = "scp -B -i %s %s %s@%s:%s" % (server.keyPath,
                    fname, server.user, server.address, server.uploadPath)
            logger.info(cmd)
            p = subprocess.Popen(args=cmd,
                        cwd=(BASE_DIR + QUEUED_DIR),
                        shell=True,
                        stdout=gw_stdout,
                        stderr=gw_stderr)
            r = p.wait()

            # Archive the file if the scp was successful
            if r == 0:
                logger.info("Sent %s" % fname)
                cmd = "/bin/mv %s %s" % (fname,
                            BASE_DIR + ARCHIVE_DIR)
                logger.debug(cmd)
                m = subprocess.Popen(args=cmd,
                        cwd=(BASE_DIR + QUEUED_DIR),
                        shell=True,
                        stdout=gw_stdout,
                        stderr=gw_stderr)
                m.wait()
            else:
                logger.error("Sending %s failed" % fname)
        else:
            logger.warn("%s too big at %s" % (fname, str(size)))


def open_ssh_tunnel(server_name, reverse_port):
    server = sshServers[server_name]
    # Determine comms window length based on batv state
    comms_window_time = COMMS_CONNECT_TIME * gwschedule.get_state_from_file()
    # Establish reverse tunnel.
    logger.info("Establishing reverse tunnel to %s:%i" % (server.address,
        reverse_port))
    cmd = "/usr/bin/ssh -Nn -i %s %s@%s -R %i:localhost:22" % (
        server.keyPath, server.user, server.address, reverse_port)
    logger.debug(cmd)
    ssh = subprocess.Popen(args=cmd,
                shell=True,
                stdout=open("/dev/null", "w"),
                stderr=gw_stderr)

    logger.info("Staying awake for %i seconds" % comms_window_time)
    time.sleep(comms_window_time)

    # Stay connected whilst comms stay awake file exists.
    while os.path.exists(COMMS_STAY_AWAKE_FILE):
        logger.info("Staying connected because of comms stay awake file")
        time.sleep(10)

    # Comms window closed, remove reverse tunnel.
    os.kill(ssh.pid, signal.SIGTERM)
    ssh.wait()

#Interface handling code


def interfaces():
    """Gets a list of all network interfaces that are up
    Grabbed from:
    http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/439093"""
    max_possible = 128  # arbitrary. raise if needed.
    bytes = max_possible * 32
    s = socket(AF_INET, SOCK_DGRAM)
    names = array.array('B', '\0' * bytes)
    outbytes = struct.unpack('iL', fcntl.ioctl(
        s.fileno(),
        0x8912,  # SIOCGIFCONF
        struct.pack('iL', bytes, names.buffer_info()[0])))[0]
    namestr = names.tostring()
    return [namestr[i:i + 32].split(
               '\0', 1)[0] for i in range(0, outbytes, 32)]


def is_interface_up(ifname):
    """checks if an interface exists and if it does if it is up"""
    if not net_exists(ifname):
        logger.error("Interface doesn't exist")
        return False

    ## {{{ http://code.activestate.com/recipes/65449/ (r1)
    # set some symbolic constants
    SIOCGIFFLAGS = 0x8913
    null256 = '\0' * 256

    # create a socket so we have a handle to query
    s = socket(AF_INET, SOCK_DGRAM)

    # call ioctl() to get the flags for the given interface
    result = fcntl.ioctl(s.fileno(), SIOCGIFFLAGS, ifname + null256)

    # extract the interface's flags from the return value
    flags, = struct.unpack('H', result[16:18])

    # check "UP" bit and print a message
    up = (flags & 1) == 1
    return up


def interface_up(ifname):
    """ starts an interface"""
    if not net_exists(ifname):
        logger.error("Interface doesn't exist so cannot up it")
        return False
        #Network exists but isn't up

    if is_interface_up(ifname):
        logger.info("Interface already up")
        interface_down(ifname)
        logger.debug("Interface taken down")
    logger.debug("ifup interface %s" % ifname)
    up = subprocess.Popen(args="/sbin/ifup %s" % ifname,
        shell=True,
        stdout=gw_stdout,
        stderr=gw_stderr)
    if up.wait() == 0:
        logger.debug("executed sucessfully")
    return is_interface_up(ifname)


def interface_down(ifname):
    if not net_exists(ifname):
        logger.error("Interface doesn't exist")
        return False
    else:
        logger.debug("ifdown interface %s" % ifname)
        down = subprocess.Popen(args="/sbin/ifdown %s" % ifname,
            shell=True,
            stdout=sys.stdout,
            stderr=sys.stderr)
        if down.wait() == 0:
            logger.debug("executed sucessfully")
        return not is_interface_up(ifname)


def net_exists(ifname):
    """Returns True if the given interface name exists"""
    return os.path.exists("/sys/class/net/%s" % ifname)


def conn_info(pid_file):
    """Returns a tuple containing:
       0) The PID of pppd
       1) The interface name.
       If there's no pppd instance, returns False"""
    if not os.path.exists(pid_file):
        return False

    try:
        f = open(pid_file)
        pid = int(f.readline().strip())
        ifname = f.readline().strip()
        return(pid, ifname)
    except:
        return False
    finally:
        f.close()


def disconnect_ppp(pid_file):
    """Disconnects and waits for pppd to terminate"""
    ci = conn_info(pid_file)

    if type(ci) != types.TupleType:
        # pppd isn't running
        return

    pid, ifname = ci

    os.kill(pid, signal.SIGTERM)

    while os.path.exists(pid_file):
        time.sleep(1)


def ppp_is_up(pid_file):
    """Returns True when the connection is up"""
    ci = conn_info(pid_file)

    if type(ci) != types.TupleType:
        return False

    pid, ifname = ci

    if ifname in interfaces():
        return True
    return False
